"""Type checker test cases"""

import os
import re
import shutil
import sys

from typing import Dict, List, Optional, Set, Tuple

from mypy import build
from mypy import defaults
from mypy.main import process_options
from mypy.test.config import test_temp_dir
from mypy.test.data import DataDrivenTestCase, DataSuite, has_stable_flags, is_incremental
from mypy.test.helpers import (
    assert_string_arrays_equal, normalize_error_messages,
    retry_on_error, testcase_pyversion, update_testcase_output,
)
from mypy.options import Options

from mypy import dmypy
from mypy import dmypy_server

# List of files that contain test case descriptions.
if sys.platform != 'win32':
    dmypy_files = [
        'check-enum.test',
        'check-incremental.test',
        'check-newtype.test',
        'check-dmypy-fine-grained.test',
    ]
else:
    dmypy_files = []  # type: List[str]


# By default we complain about missing files. This is a special module prefix
# for which we allow non-existence. This is used for testing missing files.
NON_EXISTENT_PREFIX = 'nonexistent'

# If this suffix is used together with NON_EXISTENT_PREFIX, the non-existent
# file is a .pyi file. Since the file doesn't exist, we can't automatically
# figure out the extension.
STUB_SUFFIX = '_stub'


class DmypySuite(DataSuite):
    files = dmypy_files
    base_path = test_temp_dir
    optional_out = True
    test_name_suffix = '_dmypy'

    @classmethod
    def filter(cls, testcase: DataDrivenTestCase) -> bool:
        return has_stable_flags(testcase) and is_incremental(testcase)

    def run_case(self, testcase: DataDrivenTestCase) -> None:
        assert has_stable_flags(testcase), "Testcase has varying flags"
        assert is_incremental(testcase), "Testcase is not incremental"

        # All tests run once with a cold cache, then at least once
        # with a warm cache and maybe changed files.  Expected output
        # is specified separately for each run.
        self.clear_cache()
        num_steps = max([2] + list(testcase.output2.keys()))
        # Check that there are no file changes beyond the last run (they would be ignored).
        for dn, dirs, files in os.walk(os.curdir):
            for file in files:
                m = re.search(r'\.([2-9])$', file)
                if m and int(m.group(1)) > num_steps:
                    raise ValueError(
                        'Output file {} exists though test case only has {} runs'.format(
                            file, num_steps))
        self.server = None  # type: Optional[dmypy_server.Server]
        for step in range(1, num_steps + 1):
            self.run_case_once(testcase, step)

    def clear_cache(self) -> None:
        dn = defaults.CACHE_DIR
        if os.path.exists(dn):
            shutil.rmtree(dn)

    def run_case_once(self, testcase: DataDrivenTestCase, incremental_step: int) -> None:
        assert incremental_step >= 1
        build.find_module_clear_caches()
        original_program_text = '\n'.join(testcase.input)

        if incremental_step > 1:
            # In runs 2+, copy *.[num] files to * files.
            for dn, dirs, files in os.walk(os.curdir):
                for file in files:
                    if file.endswith('.' + str(incremental_step)):
                        full = os.path.join(dn, file)
                        target = full[:-2]
                        # Use retries to work around potential flakiness on Windows (AppVeyor).
                        retry_on_error(lambda: shutil.copy(full, target))

                        # In some systems, mtime has a resolution of 1 second which can cause
                        # annoying-to-debug issues when a file has the same size after a
                        # change. We manually set the mtime to circumvent this.
                        new_time = os.stat(target).st_mtime + 1
                        os.utime(target, times=(new_time, new_time))
            # Delete files scheduled to be deleted in [delete <path>.num] sections.
            for path in testcase.deleted_paths.get(incremental_step, set()):
                # Use retries to work around potential flakiness on Windows (AppVeyor).
                retry_on_error(lambda: os.remove(path))

        module_data = self.parse_module(original_program_text, incremental_step)

        if incremental_step == 1:
            # In run 1, copy program text to program file.
            for module_name, program_path, program_text in module_data:
                if module_name == '__main__' and program_text is not None:
                    with open(program_path, 'w') as f:
                        f.write(program_text)
                    break

        # Parse options after moving files (in case mypy.ini is being moved).
        options = self.parse_options(original_program_text, testcase, incremental_step)
        if incremental_step == 1:
            server_options = []  # type: List[str]
            if 'fine-grained' in testcase.file:
                server_options.append('--experimental')
                options.fine_grained_incremental = True
                options.local_partial_types = True
            self.server = dmypy_server.Server(server_options)  # TODO: Fix ugly API
            self.server.options = options

        assert self.server is not None  # Set in step 1 and survives into next steps
        sources = []
        for module_name, program_path, program_text in module_data:
            # Always set to none so we're forced to reread the module in incremental mode
            sources.append(build.BuildSource(program_path, module_name, None))
        response = self.server.check(sources, alt_lib_path=test_temp_dir)
        a = (response['out'] or response['err']).splitlines()
        a = normalize_error_messages(a)

        # Make sure error messages match
        if incremental_step == 1:
            msg = 'Unexpected type checker output in incremental, run 1 ({}, line {})'
            output = testcase.output
        elif incremental_step > 1:
            msg = ('Unexpected type checker output in incremental, run {}'.format(
                incremental_step) + ' ({}, line {})')
            output = testcase.output2.get(incremental_step, [])
        else:
            raise AssertionError()

        if output != a and self.update_data:
            update_testcase_output(testcase, a)
        assert_string_arrays_equal(output, a, msg.format(testcase.file, testcase.line))

        manager = self.server.last_manager
        if manager is not None:
            if options.follow_imports == 'normal' and testcase.output is None:
                self.verify_cache(module_data, a, manager)
            if incremental_step > 1:
                suffix = '' if incremental_step == 2 else str(incremental_step - 1)
                self.check_module_equivalence(
                    'rechecked' + suffix,
                    testcase.expected_rechecked_modules.get(incremental_step - 1),
                    manager.rechecked_modules)
                self.check_module_equivalence(
                    'stale' + suffix,
                    testcase.expected_stale_modules.get(incremental_step - 1),
                    manager.stale_modules)

    def check_module_equivalence(self, name: str,
                                 expected: Optional[Set[str]], actual: Set[str]) -> None:
        if expected is not None:
            expected_normalized = sorted(expected)
            actual_normalized = sorted(actual.difference({"__main__"}))
            assert_string_arrays_equal(
                expected_normalized,
                actual_normalized,
                ('Actual modules ({}) do not match expected modules ({}) '
                 'for "[{} ...]"').format(
                    ', '.join(actual_normalized),
                    ', '.join(expected_normalized),
                    name))

    def verify_cache(self, module_data: List[Tuple[str, str, Optional[str]]], a: List[str],
                     manager: build.BuildManager) -> None:
        # There should be valid cache metadata for each module except
        # those in error_paths; for those there should not be.
        #
        # NOTE: When A imports B and there's an error in B, the cache
        # data for B is invalidated, but the cache data for A remains.
        # However build.process_graphs() will ignore A's cache data.
        #
        # Also note that when A imports B, and there's an error in A
        # _due to a valid change in B_, the cache data for B will be
        # invalidated and updated, but the old cache data for A will
        # remain unchanged. As before, build.process_graphs() will
        # ignore A's (old) cache data.
        error_paths = self.find_error_paths(a)
        modules = self.find_module_files()
        modules.update({module_name: path for module_name, path, text in module_data})
        missing_paths = self.find_missing_cache_files(modules, manager)
        if not missing_paths.issubset(error_paths):
            raise AssertionError("cache data discrepancy %s != %s" %
                                 (missing_paths, error_paths))

    def find_error_paths(self, a: List[str]) -> Set[str]:
        hits = set()
        for line in a:
            m = re.match(r'([^\s:]+):\d+: error:', line)
            if m:
                # Normalize to Linux paths.
                p = m.group(1).replace(os.path.sep, '/')
                hits.add(p)
        return hits

    def find_module_files(self) -> Dict[str, str]:
        modules = {}
        for dn, dirs, files in os.walk(test_temp_dir):
            dnparts = dn.split(os.sep)
            assert dnparts[0] == test_temp_dir
            del dnparts[0]
            for file in files:
                if file.endswith('.py'):
                    if file == "__init__.py":
                        # If the file path is `a/b/__init__.py`, exclude the file name
                        # and make sure the module id is just `a.b`, not `a.b.__init__`.
                        id = '.'.join(dnparts)
                    else:
                        base, ext = os.path.splitext(file)
                        id = '.'.join(dnparts + [base])
                    modules[id] = os.path.join(dn, file)
        return modules

    def find_missing_cache_files(self, modules: Dict[str, str],
                                 manager: build.BuildManager) -> Set[str]:
        ignore_errors = True
        missing = {}
        for id, path in modules.items():
            meta = build.find_cache_meta(id, path, manager)
            if not build.validate_meta(meta, id, path, ignore_errors, manager):
                missing[id] = path
        return set(missing.values())

    def parse_module(self,
                     program_text: str,
                     incremental_step: int) -> List[Tuple[str, str, Optional[str]]]:
        """Return the module and program names for a test case.

        Normally, the unit tests will parse the default ('__main__')
        module and follow all the imports listed there. You can override
        this behavior and instruct the tests to check multiple modules
        by using a comment like this in the test case input:

          # cmd: mypy -m foo.bar foo.baz

        You can also use `# cmdN:` to have a different cmd for incremental
        step N (2, 3, ...).

        Return a list of tuples (module name, file name, program text).
        """
        m = re.search('# cmd: mypy -m ([a-zA-Z0-9_. ]+)$', program_text, flags=re.MULTILINE)
        regex = '# cmd{}: mypy -m ([a-zA-Z0-9_. ]+)$'.format(incremental_step)
        alt_m = re.search(regex, program_text, flags=re.MULTILINE)
        if alt_m is not None and incremental_step > 1:
            # Optionally return a different command if in a later step
            # of incremental mode, otherwise default to reusing the
            # original cmd.
            m = alt_m

        if m:
            # The test case wants to use a non-default main
            # module. Look up the module and give it as the thing to
            # analyze.
            module_names = m.group(1)
            out = []  # type: List[Tuple[str, str, Optional[str]]]
            for module_name in module_names.split(' '):
                path = build.find_module(module_name, [test_temp_dir])
                if path is None and module_name.startswith(NON_EXISTENT_PREFIX):
                    # This is a special name for a file that we don't want to exist.
                    assert '.' not in module_name  # TODO: Packages not supported here
                    if module_name.endswith(STUB_SUFFIX):
                        fnam = '{}.pyi'.format(module_name)
                    else:
                        fnam = '{}.py'.format(module_name)
                    path = os.path.join(test_temp_dir, fnam)
                    out.append((module_name, path, None))
                else:
                    assert path is not None, "Can't find ad hoc case file for %r" % module_name
                    with open(path) as f:
                        program_text = f.read()
                    out.append((module_name, path, program_text))
            return out
        else:
            return [('__main__', 'main', program_text)]

    def parse_options(self, program_text: str, testcase: DataDrivenTestCase,
                      incremental_step: int) -> Options:
        options = Options()
        flags = re.search('# flags: (.*)$', program_text, flags=re.MULTILINE)
        if incremental_step > 1:
            flags2 = re.search('# flags{}: (.*)$'.format(incremental_step), program_text,
                               flags=re.MULTILINE)
            if flags2:
                flags = flags2

        flag_list = None
        if flags:
            flag_list = flags.group(1).split()
            targets, options = process_options(flag_list, require_targets=False)
            if targets:
                raise RuntimeError('Specifying targets via the flags pragma is not supported.')
        else:
            options = Options()

        # Allow custom python version to override testcase_pyversion
        if (not flag_list or
                all(flag not in flag_list for flag in ['--python-version', '-2', '--py2'])):
            options.python_version = testcase_pyversion(testcase.file, testcase.name)

        options.use_builtins_fixtures = True
        options.show_traceback = True
        options.incremental = True

        return options



# Stubs for termios

from typing import IO, List, Union

_FD = Union[int, IO[str]]
_Attr = List[Union[int, List[bytes]]]

# TODO constants not really documented
B0 = ...  # type: int
B1000000 = ...  # type: int
B110 = ...  # type: int
B115200 = ...  # type: int
B1152000 = ...  # type: int
B1200 = ...  # type: int
B134 = ...  # type: int
B150 = ...  # type: int
B1500000 = ...  # type: int
B1800 = ...  # type: int
B19200 = ...  # type: int
B200 = ...  # type: int
B2000000 = ...  # type: int
B230400 = ...  # type: int
B2400 = ...  # type: int
B2500000 = ...  # type: int
B300 = ...  # type: int
B3000000 = ...  # type: int
B3500000 = ...  # type: int
B38400 = ...  # type: int
B4000000 = ...  # type: int
B460800 = ...  # type: int
B4800 = ...  # type: int
B50 = ...  # type: int
B500000 = ...  # type: int
B57600 = ...  # type: int
B576000 = ...  # type: int
B600 = ...  # type: int
B75 = ...  # type: int
B921600 = ...  # type: int
B9600 = ...  # type: int
BRKINT = ...  # type: int
BS0 = ...  # type: int
BS1 = ...  # type: int
BSDLY = ...  # type: int
CBAUD = ...  # type: int
CBAUDEX = ...  # type: int
CDSUSP = ...  # type: int
CEOF = ...  # type: int
CEOL = ...  # type: int
CEOT = ...  # type: int
CERASE = ...  # type: int
CFLUSH = ...  # type: int
CIBAUD = ...  # type: int
CINTR = ...  # type: int
CKILL = ...  # type: int
CLNEXT = ...  # type: int
CLOCAL = ...  # type: int
CQUIT = ...  # type: int
CR0 = ...  # type: int
CR1 = ...  # type: int
CR2 = ...  # type: int
CR3 = ...  # type: int
CRDLY = ...  # type: int
CREAD = ...  # type: int
CRPRNT = ...  # type: int
CRTSCTS = ...  # type: int
CS5 = ...  # type: int
CS6 = ...  # type: int
CS7 = ...  # type: int
CS8 = ...  # type: int
CSIZE = ...  # type: int
CSTART = ...  # type: int
CSTOP = ...  # type: int
CSTOPB = ...  # type: int
CSUSP = ...  # type: int
CWERASE = ...  # type: int
ECHO = ...  # type: int
ECHOCTL = ...  # type: int
ECHOE = ...  # type: int
ECHOK = ...  # type: int
ECHOKE = ...  # type: int
ECHONL = ...  # type: int
ECHOPRT = ...  # type: int
EXTA = ...  # type: int
EXTB = ...  # type: int
FF0 = ...  # type: int
FF1 = ...  # type: int
FFDLY = ...  # type: int
FIOASYNC = ...  # type: int
FIOCLEX = ...  # type: int
FIONBIO = ...  # type: int
FIONCLEX = ...  # type: int
FIONREAD = ...  # type: int
FLUSHO = ...  # type: int
HUPCL = ...  # type: int
ICANON = ...  # type: int
ICRNL = ...  # type: int
IEXTEN = ...  # type: int
IGNBRK = ...  # type: int
IGNCR = ...  # type: int
IGNPAR = ...  # type: int
IMAXBEL = ...  # type: int
INLCR = ...  # type: int
INPCK = ...  # type: int
IOCSIZE_MASK = ...  # type: int
IOCSIZE_SHIFT = ...  # type: int
ISIG = ...  # type: int
ISTRIP = ...  # type: int
IUCLC = ...  # type: int
IXANY = ...  # type: int
IXOFF = ...  # type: int
IXON = ...  # type: int
NCC = ...  # type: int
NCCS = ...  # type: int
NL0 = ...  # type: int
NL1 = ...  # type: int
NLDLY = ...  # type: int
NOFLSH = ...  # type: int
N_MOUSE = ...  # type: int
N_PPP = ...  # type: int
N_SLIP = ...  # type: int
N_STRIP = ...  # type: int
N_TTY = ...  # type: int
OCRNL = ...  # type: int
OFDEL = ...  # type: int
OFILL = ...  # type: int
OLCUC = ...  # type: int
ONLCR = ...  # type: int
ONLRET = ...  # type: int
ONOCR = ...  # type: int
OPOST = ...  # type: int
PARENB = ...  # type: int
PARMRK = ...  # type: int
PARODD = ...  # type: int
PENDIN = ...  # type: int
TAB0 = ...  # type: int
TAB1 = ...  # type: int
TAB2 = ...  # type: int
TAB3 = ...  # type: int
TABDLY = ...  # type: int
TCFLSH = ...  # type: int
TCGETA = ...  # type: int
TCGETS = ...  # type: int
TCIFLUSH = ...  # type: int
TCIOFF = ...  # type: int
TCIOFLUSH = ...  # type: int
TCION = ...  # type: int
TCOFLUSH = ...  # type: int
TCOOFF = ...  # type: int
TCOON = ...  # type: int
TCSADRAIN = ...  # type: int
TCSAFLUSH = ...  # type: int
TCSANOW = ...  # type: int
TCSBRK = ...  # type: int
TCSBRKP = ...  # type: int
TCSETA = ...  # type: int
TCSETAF = ...  # type: int
TCSETAW = ...  # type: int
TCSETS = ...  # type: int
TCSETSF = ...  # type: int
TCSETSW = ...  # type: int
TCXONC = ...  # type: int
TIOCCONS = ...  # type: int
TIOCEXCL = ...  # type: int
TIOCGETD = ...  # type: int
TIOCGICOUNT = ...  # type: int
TIOCGLCKTRMIOS = ...  # type: int
TIOCGPGRP = ...  # type: int
TIOCGSERIAL = ...  # type: int
TIOCGSOFTCAR = ...  # type: int
TIOCGWINSZ = ...  # type: int
TIOCINQ = ...  # type: int
TIOCLINUX = ...  # type: int
TIOCMBIC = ...  # type: int
TIOCMBIS = ...  # type: int
TIOCMGET = ...  # type: int
TIOCMIWAIT = ...  # type: int
TIOCMSET = ...  # type: int
TIOCM_CAR = ...  # type: int
TIOCM_CD = ...  # type: int
TIOCM_CTS = ...  # type: int
TIOCM_DSR = ...  # type: int
TIOCM_DTR = ...  # type: int
TIOCM_LE = ...  # type: int
TIOCM_RI = ...  # type: int
TIOCM_RNG = ...  # type: int
TIOCM_RTS = ...  # type: int
TIOCM_SR = ...  # type: int
TIOCM_ST = ...  # type: int
TIOCNOTTY = ...  # type: int
TIOCNXCL = ...  # type: int
TIOCOUTQ = ...  # type: int
TIOCPKT = ...  # type: int
TIOCPKT_DATA = ...  # type: int
TIOCPKT_DOSTOP = ...  # type: int
TIOCPKT_FLUSHREAD = ...  # type: int
TIOCPKT_FLUSHWRITE = ...  # type: int
TIOCPKT_NOSTOP = ...  # type: int
TIOCPKT_START = ...  # type: int
TIOCPKT_STOP = ...  # type: int
TIOCSCTTY = ...  # type: int
TIOCSERCONFIG = ...  # type: int
TIOCSERGETLSR = ...  # type: int
TIOCSERGETMULTI = ...  # type: int
TIOCSERGSTRUCT = ...  # type: int
TIOCSERGWILD = ...  # type: int
TIOCSERSETMULTI = ...  # type: int
TIOCSERSWILD = ...  # type: int
TIOCSER_TEMT = ...  # type: int
TIOCSETD = ...  # type: int
TIOCSLCKTRMIOS = ...  # type: int
TIOCSPGRP = ...  # type: int
TIOCSSERIAL = ...  # type: int
TIOCSSOFTCAR = ...  # type: int
TIOCSTI = ...  # type: int
TIOCSWINSZ = ...  # type: int
TOSTOP = ...  # type: int
VDISCARD = ...  # type: int
VEOF = ...  # type: int
VEOL = ...  # type: int
VEOL2 = ...  # type: int
VERASE = ...  # type: int
VINTR = ...  # type: int
VKILL = ...  # type: int
VLNEXT = ...  # type: int
VMIN = ...  # type: int
VQUIT = ...  # type: int
VREPRINT = ...  # type: int
VSTART = ...  # type: int
VSTOP = ...  # type: int
VSUSP = ...  # type: int
VSWTC = ...  # type: int
VSWTCH = ...  # type: int
VT0 = ...  # type: int
VT1 = ...  # type: int
VTDLY = ...  # type: int
VTIME = ...  # type: int
VWERASE = ...  # type: int
XCASE = ...  # type: int
XTABS = ...  # type: int

def tcgetattr(fd: _FD) -> _Attr: ...
def tcsetattr(fd: _FD, when: int, attributes: _Attr) -> None: ...
def tcsendbreak(fd: _FD, duration: int) -> None: ...
def tcdrain(fd: _FD) -> None: ...
def tcflush(fd: _FD, queue: int) -> None: ...
def tcflow(fd: _FD, action: int) -> None: ...



"""Shared definitions used by different parts of semantic analysis."""

# Priorities for ordering of patches within the final "patch" phase of semantic analysis
# (after pass 3):

# Fix forward references (needs to happen first)
PRIORITY_FORWARD_REF = 0
# Fix fallbacks (does joins)
PRIORITY_FALLBACKS = 1
# Checks type var values (does subtype checks)
PRIORITY_TYPEVAR_VALUES = 2



"""Test cases for AST diff (used for fine-grained incremental checking)"""

import os
from typing import List, Tuple, Dict, Optional

from mypy import build
from mypy.build import BuildSource
from mypy.errors import CompileError
from mypy.nodes import MypyFile
from mypy.options import Options
from mypy.server.astdiff import snapshot_symbol_table, compare_symbol_table_snapshots
from mypy.test.config import test_temp_dir
from mypy.test.data import DataDrivenTestCase, DataSuite
from mypy.test.helpers import assert_string_arrays_equal


class ASTDiffSuite(DataSuite):
    files = ['diff.test']
    base_path = test_temp_dir
    optional_out = True

    def run_case(self, testcase: DataDrivenTestCase) -> None:
        first_src = '\n'.join(testcase.input)
        files_dict = dict(testcase.files)
        second_src = files_dict['tmp/next.py']

        messages1, files1 = self.build(first_src)
        messages2, files2 = self.build(second_src)

        a = []
        if messages1:
            a.extend(messages1)
        if messages2:
            a.append('== next ==')
            a.extend(messages2)

        assert files1 is not None and files2 is not None, ('cases where CompileError'
                                                           ' occurred should not be run')
        prefix = '__main__'
        snapshot1 = snapshot_symbol_table(prefix, files1['__main__'].names)
        snapshot2 = snapshot_symbol_table(prefix, files2['__main__'].names)
        diff = compare_symbol_table_snapshots(prefix, snapshot1, snapshot2)
        for trigger in sorted(diff):
            a.append(trigger)

        assert_string_arrays_equal(
            testcase.output, a,
            'Invalid output ({}, line {})'.format(testcase.file,
                                                  testcase.line))

    def build(self, source: str) -> Tuple[List[str], Optional[Dict[str, MypyFile]]]:
        options = Options()
        options.use_builtins_fixtures = True
        options.show_traceback = True
        options.cache_dir = os.devnull
        try:
            result = build.build(sources=[BuildSource('main', None, source)],
                                 options=options,
                                 alt_lib_path=test_temp_dir)
        except CompileError as e:
            # TODO: Is it okay to return None?
            return e.messages, None
        return result.errors, result.files



from typing import Any

COM_SLEEP = ...  # type: Any
COM_QUIT = ...  # type: Any
COM_INIT_DB = ...  # type: Any
COM_QUERY = ...  # type: Any
COM_FIELD_LIST = ...  # type: Any
COM_CREATE_DB = ...  # type: Any
COM_DROP_DB = ...  # type: Any
COM_REFRESH = ...  # type: Any
COM_SHUTDOWN = ...  # type: Any
COM_STATISTICS = ...  # type: Any
COM_PROCESS_INFO = ...  # type: Any
COM_CONNECT = ...  # type: Any
COM_PROCESS_KILL = ...  # type: Any
COM_DEBUG = ...  # type: Any
COM_PING = ...  # type: Any
COM_TIME = ...  # type: Any
COM_DELAYED_INSERT = ...  # type: Any
COM_CHANGE_USER = ...  # type: Any
COM_BINLOG_DUMP = ...  # type: Any
COM_TABLE_DUMP = ...  # type: Any
COM_CONNECT_OUT = ...  # type: Any
COM_REGISTER_SLAVE = ...  # type: Any



# NOTE: This stub is incomplete - only contains some global functions

from typing import Any, Dict, Optional

def run(statement: str,
        globals: Optional[Dict[str, Any]] = ...,
        locals: Optional[Dict[str, Any]] = ...) -> None:
    ...

def runeval(expression: str,
            globals: Optional[Dict[str, Any]] = ...,
            locals: Optional[Dict[str, Any]] = ...) -> Any:
    ...

def runctx(statement: str,
        globals: Dict[str, Any],
        locals: Dict[str, Any]) -> None:
    ...

def runcall(*args: Any, **kwds: Any) -> Any:
    ...

def set_trace() -> None:
    ...

def post_mortem(t: Optional[Any] = ...) -> None:
    ...

def pm() -> None:
    ...



"""Plugin for supporting the attrs library (http://www.attrs.org)"""
from collections import OrderedDict
from typing import Optional, Dict, List, cast, Tuple, Iterable

import mypy.plugin  # To avoid circular imports.
from mypy.exprtotype import expr_to_unanalyzed_type, TypeTranslationError
from mypy.fixup import lookup_qualified_stnode
from mypy.nodes import (
    Context, Argument, Var, ARG_OPT, ARG_POS, TypeInfo, AssignmentStmt,
    TupleExpr, ListExpr, NameExpr, CallExpr, RefExpr, FuncBase,
    is_class_var, TempNode, Decorator, MemberExpr, Expression, FuncDef, Block,
    PassStmt, SymbolTableNode, MDEF, JsonDict
)
from mypy.types import (
    Type, AnyType, TypeOfAny, CallableType, NoneTyp, TypeVarDef, TypeVarType,
    Overloaded, Instance
)
from mypy.typevars import fill_typevars


# The names of the different functions that create classes or arguments.
attr_class_makers = {
    'attr.s',
    'attr.attrs',
    'attr.attributes',
}
attr_dataclass_makers = {
    'attr.dataclass',
}
attr_attrib_makers = {
    'attr.ib',
    'attr.attrib',
    'attr.attr',
}


class Attribute:
    """The value of an attr.ib() call."""

    def __init__(self, name: str, info: TypeInfo,
                 has_default: bool, init: bool, converter_name: Optional[str],
                 context: Context) -> None:
        self.name = name
        self.info = info
        self.has_default = has_default
        self.init = init
        self.converter_name = converter_name
        self.context = context

    def argument(self, ctx: 'mypy.plugin.ClassDefContext') -> Argument:
        """Return this attribute as an argument to __init__."""
        assert self.init
        init_type = self.info[self.name].type

        if self.converter_name:
            # When a converter is set the init_type is overriden by the first argument
            # of the converter method.
            converter = lookup_qualified_stnode(ctx.api.modules, self.converter_name, True)
            if not converter:
                # The converter may be a local variable. Check there too.
                converter = ctx.api.lookup_qualified(self.converter_name, self.info, True)

            if (converter
                    and converter.type
                    and isinstance(converter.type, CallableType)
                    and converter.type.arg_types):
                init_type = ctx.api.anal_type(converter.type.arg_types[0])
            else:
                ctx.api.fail("Cannot determine type of converter function", self.context)
                init_type = AnyType(TypeOfAny.from_error)
        elif self.converter_name == '':
            # This means we had a converter but it's not of a type we can infer.
            # Error was shown in _get_converter_name
            init_type = AnyType(TypeOfAny.from_error)

        if init_type is None:
            if ctx.api.options.disallow_untyped_defs:
                # This is a compromise.  If you don't have a type here then the
                # __init__ will be untyped. But since the __init__ is added it's
                # pointing at the decorator. So instead we also show the error in the
                # assignment, which is where you would fix the issue.
                node = self.info[self.name].node
                assert node is not None
                ctx.api.msg.need_annotation_for_var(node, self.context)

            # Convert type not set to Any.
            init_type = AnyType(TypeOfAny.unannotated)

        # Attrs removes leading underscores when creating the __init__ arguments.
        return Argument(Var(self.name.lstrip("_"), init_type), init_type,
                        None,
                        ARG_OPT if self.has_default else ARG_POS)

    def serialize(self) -> JsonDict:
        """Serialize this object so it can be saved and restored."""
        return {
            'name': self.name,
            'has_default': self.has_default,
            'init': self.init,
            'converter_name': self.converter_name,
            'context_line': self.context.line,
            'context_column': self.context.column,
        }

    @classmethod
    def deserialize(cls, info: TypeInfo, data: JsonDict) -> 'Attribute':
        """Return the Attribute that was serialized."""
        return Attribute(
            data['name'],
            info,
            data['has_default'],
            data['init'],
            data['converter_name'],
            Context(line=data['context_line'], column=data['context_column'])
        )


def attr_class_maker_callback(ctx: 'mypy.plugin.ClassDefContext',
                              auto_attribs_default: bool = False) -> None:
    """Add necessary dunder methods to classes decorated with attr.s.

    attrs is a package that lets you define classes without writing dull boilerplate code.

    At a quick glance, the decorator searches the class body for assignments of `attr.ib`s (or
    annotated variables if auto_attribs=True), then depending on how the decorator is called,
    it will add an __init__ or all the __cmp__ methods.  For frozen=True it will turn the attrs
    into properties.

    See http://www.attrs.org/en/stable/how-does-it-work.html for information on how attrs works.
    """
    info = ctx.cls.info

    init = _get_decorator_bool_argument(ctx, 'init', True)
    frozen = _get_decorator_bool_argument(ctx, 'frozen', False)
    cmp = _get_decorator_bool_argument(ctx, 'cmp', True)
    auto_attribs = _get_decorator_bool_argument(ctx, 'auto_attribs', auto_attribs_default)

    if ctx.api.options.python_version[0] < 3:
        if auto_attribs:
            ctx.api.fail("auto_attribs is not supported in Python 2", ctx.reason)
            return
        if not info.defn.base_type_exprs:
            # Note: This will not catch subclassing old-style classes.
            ctx.api.fail("attrs only works with new-style classes", info.defn)
            return

    attributes = _analyze_class(ctx, auto_attribs)

    adder = MethodAdder(info, ctx.api.named_type('__builtins__.function'))
    if init:
        _add_init(ctx, attributes, adder)
    if cmp:
        _add_cmp(ctx, adder)
    if frozen:
        _make_frozen(ctx, attributes)


def _analyze_class(ctx: 'mypy.plugin.ClassDefContext', auto_attribs: bool) -> List[Attribute]:
    """Analyze the class body of an attr maker, its parents, and return the Attributes found."""
    own_attrs = OrderedDict()  # type: OrderedDict[str, Attribute]
    # Walk the body looking for assignments and decorators.
    for stmt in ctx.cls.defs.body:
        if isinstance(stmt, AssignmentStmt):
            for attr in _attributes_from_assignment(ctx, stmt, auto_attribs):
                # When attrs are defined twice in the same body we want to use the 2nd definition
                # in the 2nd location. So remove it from the OrderedDict.
                # Unless it's auto_attribs in which case we want the 2nd definition in the
                # 1st location.
                if not auto_attribs and attr.name in own_attrs:
                    del own_attrs[attr.name]
                own_attrs[attr.name] = attr
        elif isinstance(stmt, Decorator):
            _cleanup_decorator(stmt, own_attrs)

    # Traverse the MRO and collect attributes from the parents.
    taken_attr_names = set(own_attrs)
    super_attrs = []
    for super_info in ctx.cls.info.mro[1:-1]:
        if 'attrs' in super_info.metadata:
            for data in super_info.metadata['attrs']['attributes']:
                # Only add an attribute if it hasn't been defined before.  This
                # allows for overwriting attribute definitions by subclassing.
                if data['name'] not in taken_attr_names:
                    a = Attribute.deserialize(super_info, data)
                    super_attrs.append(a)
                    taken_attr_names.add(a.name)
    attributes = super_attrs + list(own_attrs.values())

    # Save the attributes so that subclasses can reuse them.
    ctx.cls.info.metadata['attrs'] = {'attributes': [attr.serialize() for attr in attributes]}

    # Check the init args for correct default-ness.  Note: This has to be done after all the
    # attributes for all classes have been read, because subclasses can override parents.
    last_default = False
    for attribute in attributes:
        if attribute.init and not attribute.has_default and last_default:
            ctx.api.fail(
                "Non-default attributes not allowed after default attributes.",
                attribute.context)
        last_default = attribute.has_default

    return attributes


def _attributes_from_assignment(ctx: 'mypy.plugin.ClassDefContext',
                                stmt: AssignmentStmt, auto_attribs: bool) -> Iterable[Attribute]:
    """Return Attribute objects that are created by this assignment.

    The assignments can look like this:
        x = attr.ib()
        x = y = attr.ib()
        x, y = attr.ib(), attr.ib()
    or if auto_attribs is enabled also like this:
        x: type
        x: type = default_value
    """
    for lvalue in stmt.lvalues:
        lvalues, rvalues = _parse_assignments(lvalue, stmt)

        if len(lvalues) != len(rvalues):
            # This means we have some assignment that isn't 1 to 1.
            # It can't be an attrib.
            continue

        for lhs, rvalue in zip(lvalues, rvalues):
            # Check if the right hand side is a call to an attribute maker.
            if (isinstance(rvalue, CallExpr)
                    and isinstance(rvalue.callee, RefExpr)
                    and rvalue.callee.fullname in attr_attrib_makers):
                attr = _attribute_from_attrib_maker(ctx, auto_attribs, lhs, rvalue, stmt)
                if attr:
                    yield attr
            elif auto_attribs and stmt.type and stmt.new_syntax and not is_class_var(lhs):
                yield _attribute_from_auto_attrib(ctx, lhs, rvalue, stmt)


def _cleanup_decorator(stmt: Decorator, attr_map: Dict[str, Attribute]) -> None:
    """Handle decorators in class bodies.

    `x.default` will set a default value on x
    `x.validator` and `x.default` will get removed to avoid throwing a type error.
    """
    remove_me = []
    for func_decorator in stmt.decorators:
        if (isinstance(func_decorator, MemberExpr)
                and isinstance(func_decorator.expr, NameExpr)
                and func_decorator.expr.name in attr_map):

            if func_decorator.name == 'default':
                attr_map[func_decorator.expr.name].has_default = True

            if func_decorator.name in ('default', 'validator'):
                # These are decorators on the attrib object that only exist during
                # class creation time.  In order to not trigger a type error later we
                # just remove them.  This might leave us with a Decorator with no
                # decorators (Emperor's new clothes?)
                # TODO: It would be nice to type-check these rather than remove them.
                #       default should be Callable[[], T]
                #       validator should be Callable[[Any, 'Attribute', T], Any]
                #       where T is the type of the attribute.
                remove_me.append(func_decorator)
    for dec in remove_me:
        stmt.decorators.remove(dec)


def _attribute_from_auto_attrib(ctx: 'mypy.plugin.ClassDefContext',
                                lhs: NameExpr,
                                rvalue: Expression,
                                stmt: AssignmentStmt) -> Attribute:
    """Return an Attribute for a new type assignment."""
    # `x: int` (without equal sign) assigns rvalue to TempNode(AnyType())
    has_rhs = not isinstance(rvalue, TempNode)
    return Attribute(lhs.name, ctx.cls.info, has_rhs, True, None, stmt)


def _attribute_from_attrib_maker(ctx: 'mypy.plugin.ClassDefContext',
                                 auto_attribs: bool,
                                 lhs: NameExpr,
                                 rvalue: CallExpr,
                                 stmt: AssignmentStmt) -> Optional[Attribute]:
    """Return an Attribute from the assignment or None if you can't make one."""
    if auto_attribs and not stmt.new_syntax:
        # auto_attribs requires an annotation on *every* attr.ib.
        assert lhs.node is not None
        ctx.api.msg.need_annotation_for_var(lhs.node, stmt)
        return None

    if len(stmt.lvalues) > 1:
        ctx.api.fail("Too many names for one attribute", stmt)
        return None

    # This is the type that belongs in the __init__ method for this attrib.
    init_type = stmt.type

    # Read all the arguments from the call.
    init = _get_bool_argument(ctx, rvalue, 'init', True)
    # TODO: Check for attr.NOTHING
    attr_has_default = bool(_get_argument(rvalue, 'default'))

    # If the type isn't set through annotation but is passed through `type=` use that.
    type_arg = _get_argument(rvalue, 'type')
    if type_arg and not init_type:
        try:
            un_type = expr_to_unanalyzed_type(type_arg)
        except TypeTranslationError:
            ctx.api.fail('Invalid argument to type', type_arg)
        else:
            init_type = ctx.api.anal_type(un_type)
            if init_type and isinstance(lhs.node, Var) and not lhs.node.type:
                # If there is no annotation, add one.
                lhs.node.type = init_type
                lhs.is_inferred_def = False

    # Note: convert is deprecated but works the same as converter.
    converter = _get_argument(rvalue, 'converter')
    convert = _get_argument(rvalue, 'convert')
    if convert and converter:
        ctx.api.fail("Can't pass both `convert` and `converter`.", rvalue)
    elif convert:
        ctx.api.fail("convert is deprecated, use converter", rvalue)
        converter = convert
    converter_name = _get_converter_name(ctx, converter)

    return Attribute(lhs.name, ctx.cls.info, attr_has_default, init, converter_name, stmt)


def _get_converter_name(ctx: 'mypy.plugin.ClassDefContext',
                        converter: Optional[Expression]) -> Optional[str]:
    """Return the full name of the converter if it exists and is a simple function."""
    # TODO: Support complex converters, e.g. lambdas, calls, etc.
    if converter:
        if (isinstance(converter, RefExpr)
                and converter.node
                and isinstance(converter.node, FuncBase)
                and converter.node.type
                and isinstance(converter.node.type, CallableType)
                and converter.node.type.arg_types):
            return converter.node.fullname()
        # Signal that we have an unsupported converter.
        ctx.api.fail(
            "Unsupported converter function, only named functions are currently supported",
            converter
        )
        return ''
    return None


def _parse_assignments(
        lvalue: Expression,
        stmt: AssignmentStmt) -> Tuple[List[NameExpr], List[Expression]]:
    """Convert a possibly complex assignment expression into lists of lvalues and rvalues."""
    lvalues = []  # type: List[NameExpr]
    rvalues = []  # type: List[Expression]
    if isinstance(lvalue, (TupleExpr, ListExpr)):
        if all(isinstance(item, NameExpr) for item in lvalue.items):
            lvalues = cast(List[NameExpr], lvalue.items)
        if isinstance(stmt.rvalue, (TupleExpr, ListExpr)):
            rvalues = stmt.rvalue.items
    elif isinstance(lvalue, NameExpr):
        lvalues = [lvalue]
        rvalues = [stmt.rvalue]
    return lvalues, rvalues


def _add_cmp(ctx: 'mypy.plugin.ClassDefContext', adder: 'MethodAdder') -> None:
    """Generate all the cmp methods for this class."""
    # For __ne__ and __eq__ the type is:
    #     def __ne__(self, other: object) -> bool
    bool_type = ctx.api.named_type('__builtins__.bool')
    object_type = ctx.api.named_type('__builtins__.object')
    args = [Argument(Var('other', object_type), object_type, None, ARG_POS)]
    for method in ['__ne__', '__eq__']:
        adder.add_method(method, args, bool_type)
    # For the rest we use:
    #    AT = TypeVar('AT')
    #    def __lt__(self: AT, other: AT) -> bool
    # This way comparisons with subclasses will work correctly.
    tvd = TypeVarDef('AT', 'AT', 1, [], object_type)
    tvd_type = TypeVarType(tvd)
    args = [Argument(Var('other', tvd_type), tvd_type, None, ARG_POS)]
    for method in ['__lt__', '__le__', '__gt__', '__ge__']:
        adder.add_method(method, args, bool_type, self_type=tvd_type, tvd=tvd)


def _make_frozen(ctx: 'mypy.plugin.ClassDefContext', attributes: List[Attribute]) -> None:
    """Turn all the attributes into properties to simulate frozen classes."""
    # TODO: Handle subclasses of frozen classes.
    for attribute in attributes:
        node = ctx.cls.info.names[attribute.name].node
        assert isinstance(node, Var)
        node.is_initialized_in_class = False
        node.is_property = True


def _add_init(ctx: 'mypy.plugin.ClassDefContext', attributes: List[Attribute],
              adder: 'MethodAdder') -> None:
    """Generate an __init__ method for the attributes and add it to the class."""
    adder.add_method(
        '__init__',
        [attribute.argument(ctx) for attribute in attributes if attribute.init],
        NoneTyp()
    )
    for stmt in ctx.cls.defs.body:
        # The type of classmethods will be wrong because it's based on the parent's __init__.
        # Set it correctly.
        if isinstance(stmt, Decorator) and stmt.func.is_class:
            func_type = stmt.func.type
            if isinstance(func_type, CallableType):
                func_type.arg_types[0] = ctx.api.class_type(ctx.cls.info)


def _get_decorator_bool_argument(
        ctx: 'mypy.plugin.ClassDefContext',
        name: str,
        default: bool) -> bool:
    """Return the bool argument for the decorator.

    This handles both @attr.s(...) and @attr.s
    """
    if isinstance(ctx.reason, CallExpr):
        return _get_bool_argument(ctx, ctx.reason, name, default)
    else:
        return default


def _get_bool_argument(ctx: 'mypy.plugin.ClassDefContext', expr: CallExpr,
                       name: str, default: bool) -> bool:
    """Return the boolean value for an argument to a call or the default if it's not found."""
    attr_value = _get_argument(expr, name)
    if attr_value:
        ret = ctx.api.parse_bool(attr_value)
        if ret is None:
            ctx.api.fail('"{}" argument must be True or False.'.format(name), expr)
            return default
        return ret
    return default


def _get_argument(call: CallExpr, name: str) -> Optional[Expression]:
    """Return the expression for the specific argument."""
    # To do this we use the CallableType of the callee to find the FormalArgument,
    # then walk the actual CallExpr looking for the appropriate argument.
    #
    # Note: I'm not hard-coding the index so that in the future we can support other
    # attrib and class makers.
    callee_type = None
    if (isinstance(call.callee, RefExpr)
            and isinstance(call.callee.node, (Var, FuncBase))
            and call.callee.node.type):
        callee_node_type = call.callee.node.type
        if isinstance(callee_node_type, Overloaded):
            # We take the last overload.
            callee_type = callee_node_type.items()[-1]
        elif isinstance(callee_node_type, CallableType):
            callee_type = callee_node_type

    if not callee_type:
        return None

    argument = callee_type.argument_by_name(name)
    if not argument:
        return None
    assert argument.name

    for i, (attr_name, attr_value) in enumerate(zip(call.arg_names, call.args)):
        if argument.pos is not None and not attr_name and i == argument.pos:
            return attr_value
        if attr_name == argument.name:
            return attr_value
    return None


class MethodAdder:
    """Helper to add methods to a TypeInfo.

    info: The TypeInfo on which we will add methods.
    function_type: The type of __builtins__.function that will be used as the
                   fallback for all methods added.
    """

    # TODO: Combine this with the code build_namedtuple_typeinfo to support both.

    def __init__(self, info: TypeInfo, function_type: Instance) -> None:
        self.info = info
        self.self_type = fill_typevars(info)
        self.function_type = function_type

    def add_method(self,
                   method_name: str, args: List[Argument], ret_type: Type,
                   self_type: Optional[Type] = None,
                   tvd: Optional[TypeVarDef] = None) -> None:
        """Add a method: def <method_name>(self, <args>) -> <ret_type>): ... to info.

        self_type: The type to use for the self argument or None to use the inferred self type.
        tvd: If the method is generic these should be the type variables.
        """
        from mypy.semanal import set_callable_name
        self_type = self_type if self_type is not None else self.self_type
        args = [Argument(Var('self'), self_type, None, ARG_POS)] + args
        arg_types = [arg.type_annotation for arg in args]
        arg_names = [arg.variable.name() for arg in args]
        arg_kinds = [arg.kind for arg in args]
        assert None not in arg_types
        signature = CallableType(cast(List[Type], arg_types), arg_kinds, arg_names,
                                 ret_type, self.function_type)
        if tvd:
            signature.variables = [tvd]
        func = FuncDef(method_name, args, Block([PassStmt()]))
        func.info = self.info
        func.type = set_callable_name(signature, func)
        func._fullname = self.info.fullname() + '.' + method_name
        func.line = self.info.line
        self.info.names[method_name] = SymbolTableNode(MDEF, func)
        # Add the created methods to the body so that they can get further semantic analysis.
        # e.g. Forward Reference Resolution.
        self.info.defn.defs.body.append(func)



# Stubs for zipapp (Python 3.5+)

from pathlib import Path
from typing import BinaryIO, Optional, Union

_Path = Union[str, Path, BinaryIO]

class ZipAppError(Exception): ...

def create_archive(source: _Path, target: Optional[_Path] = ..., interpreter: Optional[str] = ..., main: Optional[str] = ...) -> None: ...
def get_interpreter(archive: _Path) -> str: ...



from typing import overload, Any, List, Optional, SupportsFloat, TypeVar, Union
from datetime import date, datetime, timedelta

__all__ = ...  # type: List[str]

_SelfT = TypeVar('_SelfT', bound=relativedelta)
_DateT = TypeVar('_DateT', date, datetime)


class weekday(object):
    def __init__(self, weekday: int, n: Optional[int]=...) -> None: ...

    def __call__(self, n: int) -> 'weekday': ...

    def __eq__(self, other) -> bool: ...

    def __repr__(self) -> str: ...

    weekday = ...  # type: int
    n = ...  # type: int

MO = ...  # type: weekday
TU = ...  # type: weekday
WE = ...  # type: weekday
TH = ...  # type: weekday
FR = ...  # type: weekday
SA = ...  # type: weekday
SU = ...  # type: weekday


class relativedelta(object):
    def __init__(self,
                 dt1: Optional[date]=...,
                 dt2: Optional[date]=...,
                 years: Optional[int]=..., months: Optional[int]=...,
                 days: Optional[int]=..., leapdays: Optional[int]=...,
                 weeks: Optional[int]=...,
                 hours: Optional[int]=..., minutes: Optional[int]=...,
                 seconds: Optional[int]=..., microseconds: Optional[int]=...,
                 year: Optional[int]=..., month: Optional[int]=...,
                 day: Optional[int]=...,
                 weekday: Optional[Union[int, weekday]]=...,
                 yearday: Optional[int]=...,
                 nlyearday: Optional[int]=...,
                 hour: Optional[int]=..., minute: Optional[int]=...,
                 second: Optional[int]=...,
                 microsecond: Optional[int]=...) -> None: ...
    @property
    def weeks(self) -> int: ...
    @weeks.setter
    def weeks(self, value: int) -> None: ...
    def normalized(self: _SelfT) -> _SelfT: ...
    # TODO: use Union when mypy will handle it properly in overloaded operator
    # methods (#2129, #1442, #1264 in mypy)
    @overload
    def __add__(self: _SelfT, other: relativedelta) -> _SelfT: ...
    @overload
    def __add__(self: _SelfT, other: timedelta) -> _SelfT: ...
    @overload
    def __add__(self, other: _DateT) -> _DateT: ...
    @overload
    def __radd__(self: _SelfT, other: relativedelta) -> _SelfT: ...
    @overload
    def __radd__(self: _SelfT, other: timedelta) -> _SelfT: ...
    @overload
    def __radd__(self, other: _DateT) -> _DateT: ...
    @overload
    def __rsub__(self: _SelfT, other: relativedelta) -> _SelfT: ...
    @overload
    def __rsub__(self: _SelfT, other: timedelta) -> _SelfT: ...
    @overload
    def __rsub__(self, other: _DateT) -> _DateT: ...
    def __sub__(self: _SelfT, other: relativedelta) -> _SelfT: ...
    def __neg__(self: _SelfT) -> _SelfT: ...
    def __bool__(self) -> bool: ...
    def __nonzero__(self) -> bool: ...
    def __mul__(self: _SelfT, other: SupportsFloat) -> _SelfT: ...
    def __rmul__(self: _SelfT, other: SupportsFloat) -> _SelfT: ...
    def __eq__(self, other) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __div__(self: _SelfT, other: SupportsFloat) -> _SelfT: ...
    def __truediv__(self: _SelfT, other: SupportsFloat) -> _SelfT: ...
    def __repr__(self) -> str: ...



from typing import Optional, Dict, Union
from mypy.types import TypeVarDef, TypeVarId
from mypy.nodes import TypeVarExpr, SymbolTableNode


class TypeVarScope:
    """Scope that holds bindings for type variables. Node fullname -> TypeVarDef."""

    def __init__(self,
                 parent: 'Optional[TypeVarScope]' = None,
                 is_class_scope: bool = False,
                 prohibited: 'Optional[TypeVarScope]' = None) -> None:
        """Initializer for TypeVarScope

        Parameters:
          parent: the outer scope for this scope
          is_class_scope: True if this represents a generic class
          prohibited: Type variables that aren't strictly in scope exactly,
                      but can't be bound because they're part of an outer class's scope.
        """
        self.scope = {}  # type: Dict[str, TypeVarDef]
        self.parent = parent
        self.func_id = 0
        self.class_id = 0
        self.is_class_scope = is_class_scope
        self.prohibited = prohibited
        if parent is not None:
            self.func_id = parent.func_id
            self.class_id = parent.class_id

    def get_function_scope(self) -> 'Optional[TypeVarScope]':
        """Get the nearest parent that's a function scope, not a class scope"""
        it = self  # type: Optional[TypeVarScope]
        while it is not None and it.is_class_scope:
            it = it.parent
        return it

    def allow_binding(self, fullname: str) -> bool:
        if fullname in self.scope:
            return False
        elif self.parent and not self.parent.allow_binding(fullname):
            return False
        elif self.prohibited and not self.prohibited.allow_binding(fullname):
            return False
        return True

    def method_frame(self) -> 'TypeVarScope':
        """A new scope frame for binding a method"""
        return TypeVarScope(self, False, None)

    def class_frame(self) -> 'TypeVarScope':
        """A new scope frame for binding a class. Prohibits *this* class's tvars"""
        return TypeVarScope(self.get_function_scope(), True, self)

    def bind_new(self, name: str, tvar_expr: TypeVarExpr) -> TypeVarDef:
        if self.is_class_scope:
            self.class_id += 1
            i = self.class_id
        else:
            self.func_id -= 1
            i = self.func_id
        tvar_def = TypeVarDef(name,
                              tvar_expr.fullname(),
                              i,
                              values=tvar_expr.values,
                              upper_bound=tvar_expr.upper_bound,
                              variance=tvar_expr.variance,
                              line=tvar_expr.line,
                              column=tvar_expr.column)
        self.scope[tvar_expr.fullname()] = tvar_def
        return tvar_def

    def bind_existing(self, tvar_def: TypeVarDef) -> None:
        self.scope[tvar_def.fullname] = tvar_def

    def get_binding(self, item: Union[str, SymbolTableNode]) -> Optional[TypeVarDef]:
        fullname = item.fullname if isinstance(item, SymbolTableNode) else item
        assert fullname is not None
        if fullname in self.scope:
            return self.scope[fullname]
        elif self.parent is not None:
            return self.parent.get_binding(fullname)
        else:
            return None

    def __str__(self) -> str:
        me = ", ".join('{}: {}`{}'.format(k, v.name, v.id) for k, v in self.scope.items())
        if self.parent is None:
            return me
        return "{} <- {}".format(str(self.parent), me)
